@@ -6,7 +6,7 @@
  *
  * Created 9 May 1998
  * Bert Bos <bert@w3.org>
- * $Id: hxnormalize.c,v 1.20 2016/03/10 17:34:16 bbos Exp $
+ * $Id: hxnormalize.c,v 1.21 2016/10/17 17:37:28 bbos Exp $
  */
 #include "config.h"
 #include <stdio.h>
@@ -255,6 +255,7 @@ static void pp(Tree n, bool preformatted, bool allow_text,
       }
       if (is_empty(n->name)) {
 	assert(n->children == NULL);
+	outbreakpoint();
 	out(do_xml ? " />" : ">", true);
 	if (break_before(n->name)) dec_indent();
 	if (!preformatted && break_after(n->name)) outln(NULL, false);
@@ -263,6 +264,7 @@ static void pp(Tree n, bool preformatted, bool allow_text,
 	/* Insert <![CDATA[...]]>, but only if input was HTML, not XML */
 	if (!n->children) {
 	  out(" />", true);
+	  if (break_before(n->name)) dec_indent();
 	} else {
 	  out(">", preformatted);
 	  /* TODO: Strictly speaking, if the input is HTML (not XML),
@@ -277,8 +279,10 @@ static void pp(Tree n, bool preformatted, bool allow_text,
 	    out(l->text, true);
 	  }
 	  if (!hasprefix(n->children->text, "<![CDATA[")) out("]]>", true);
+	  if (break_before(n->name)) dec_indent();
 	  out("</", preformatted);
 	  out(n->name, preformatted);
+	  outbreakpoint();
 	  out(">", preformatted);
 	}
 	if (!preformatted && break_after(n->name)) outbreak();
@@ -292,12 +296,15 @@ static void pp(Tree n, bool preformatted, bool allow_text,
 	s = n->children->text + 9; /* Skip "<![CDATA[" */
 	i = strlen(s) - 3;	   /* Omit "]]>" */
 	for (j = 0; j < i; j++) outc(s[j], true);
+	if (break_before(n->name)) dec_indent();
 	out("</", preformatted);
 	out(n->name, preformatted);
+	outbreakpoint();
 	out(">", preformatted);
 	if (!preformatted && break_after(n->name)) outbreak();
 
       } else {
+	outbreakpoint();
 	out(">", preformatted);
 	pre = preformatted || is_pre(n->name);
 	mixed = is_mixed(n->name);
@@ -305,7 +312,8 @@ static void pp(Tree n, bool preformatted, bool allow_text,
 	  pp(l, pre, mixed, lang);
 	if (break_before(n->name)) dec_indent();
 	if (do_xml || do_endtag || need_etag(n->name) || next_ambiguous(n)) {
-	  out("</", preformatted); out(n->name, preformatted);
+	  out("</", pre); out(n->name, pre);
+	  outbreakpoint();
 	  out(">", preformatted);
 	}
 	if (!preformatted && break_after(n->name)) outbreak();
