@@ -28,6 +28,13 @@
 #endif
 #include <ctype.h>
 #include <regex.h>
+#include <err.h>
+#include <sysexits.h>
+#if HAVE_LIBIDN2
+# include <idn2.h>
+#elif HAVE_LIBIDN
+# include <idna.h>
+#endif
 #include "export.h"
 #include "heap.e"
 #include "types.e"
@@ -45,6 +52,35 @@ EXPORT typedef struct {
 } *URL;
 
 
+/* idn_to_ascii -- convert UTF-8 to Punycode, allocate on heap */
+static string idn_to_ascii(const conststring s)
+{
+#if HAVE_LIBIDN2
+  string p;
+  int rc;
+
+  rc = idn2_lookup_u8(s, (uint8_t**)&p, IDN2_NFC_INPUT);
+  if (rc == IDN2_OK) return p;
+  warnx(idn2_strerror(rc));
+  return newstring(s);
+#elif HAVE_LIBIDN
+  string p;
+  Idna_rc rc;
+
+  rc = idna_to_ascii_8z(s, &p, 0);
+  if (rc == IDNA_SUCCESS) return p;
+  warnx("%s", idna_strerror(rc));
+  return newstring(s);
+#else
+  conststring p;
+
+  for (p = s; *p; p++)
+    if ((unsigned char)(*p) < 33 || (unsigned char)(*p) > 127)
+      warnx("Internationalized domain names in URLs require libidn.");
+  return newstring(s);
+#endif
+}
+
 /* utf8tohex -- convert UTF-8 to %HH hex encoding, allocate on heap */
 static string utf8tohex(const conststring s)
 {
@@ -52,10 +88,13 @@ static string utf8tohex(const conststring s)
   string h;
   int i, j;
 
+  if (!s) return NULL;
   newarray(h, 3 * strlen(s) + 1);		/* Usually too much */
   for (i = 0, j = 0; s[i]; i++) {
     if (s[i] & 0x80) {				/* Not ASCII */
-      h[j++] = '%'; h[j++] = hex[s[i]/16]; h[j++] = hex[s[i]%16];
+      h[j++] = '%';
+      h[j++] = hex[((unsigned char)s[i])/16];
+      h[j++] = hex[((unsigned char)s[i])%16];
     } else if (s[i] == ' ') {			/* Also escape spaces */
       h[j++] = '%'; h[j++] = '2'; h[j++] = '0';
     } else
@@ -130,7 +169,7 @@ EXPORT URL URL_new(const conststring url)
 
   /* Store the various parts */
   new(result);
-  result->full = utf8tohex(url);
+  result->full = newstring(url);
   result->proto = pm[2].rm_so == -1
     ? NULL : down(newnstring(url, pm[2].rm_eo));
   result->user = pm[5].rm_so == -1
@@ -224,7 +263,6 @@ static void remove_dot_segments(string path)
 EXPORT URL URL_absolutize(const URL base, const URL url)
 {
   URL abs;
-  string s;
 
   new(abs);
 
@@ -274,7 +312,7 @@ EXPORT URL URL_absolutize(const URL base, const URL url)
   }
   abs->fragment = newstring(url->fragment);
 
-  newarray(s, (abs->proto ? strlen(abs->proto) + 1 : 0)
+  newarray(abs->full, (abs->proto ? strlen(abs->proto) + 1 : 0)
 	   + (abs->user ? strlen(abs->user) + 1 : 0)
 	   + (abs->password ? strlen(abs->password) + 1 : 0)
 	   + (abs->machine ? strlen(abs->machine) + 4 : 0)
@@ -283,7 +321,7 @@ EXPORT URL URL_absolutize(const URL base, const URL url)
 	   + (abs->query ? strlen(abs->query) + 1 : 0)
 	   + (abs->fragment ? strlen(abs->fragment) + 1 : 0)
 	   + 1);
-  sprintf(s, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
+  sprintf(abs->full, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
 	  abs->proto ? abs->proto : (string) "",
 	  abs->proto ? (string) ":" : (string) "",
 	  abs->machine ? (string) "//" : (string) "",
@@ -302,8 +340,6 @@ EXPORT URL URL_absolutize(const URL base, const URL url)
 	  abs->fragment ? (string) "#" : (string) "",
 	  abs->fragment ? abs->fragment : (string) "");
   /* Instead of strchr() above, we could have an IPv6 flag. Necessary? */
-  abs->full = utf8tohex(s);
-  dispose(s);
 
   return abs;
 }
@@ -319,3 +355,61 @@ EXPORT string URL_s_absolutize(const conststring base, const conststring url)
   URL_dispose(base1);
   return result;
 }
+
+/* URL_to_ascii -- use punycode and %-escaping to turn an IRI into a URI */
+EXPORT URL URL_to_ascii(const URL iri)
+{
+  URL url;
+
+  new(url);
+  url->proto = newstring(iri->proto);
+  url->user = newstring(iri->user);
+  url->password = newstring(iri->password);
+  url->machine = idn_to_ascii(iri->machine);
+  url->port = newstring(iri->port);
+  url->path = utf8tohex(iri->path);
+  url->query = utf8tohex(iri->query);
+  url->fragment = utf8tohex(iri->fragment);
+
+  newarray(url->full, (url->proto ? strlen(url->proto) + 1 : 0)
+	   + (url->user ? strlen(url->user) + 1 : 0)
+	   + (url->password ? strlen(url->password) + 1 : 0)
+	   + (url->machine ? strlen(url->machine) + 4 : 0)
+	   + (url->port ? strlen(url->port) + 1 : 0)
+	   + (url->path ? strlen(url->path) : 0)
+	   + (url->query ? strlen(url->query) + 1 : 0)
+	   + (url->fragment ? strlen(url->fragment) + 1 : 0)
+	   + 1);
+  sprintf(url->full, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
+	  url->proto ? url->proto : (string) "",
+	  url->proto ? (string) ":" : (string) "",
+	  url->machine ? (string) "//" : (string) "",
+	  url->user ? url->user : (string) "",
+	  url->password ? (string) ":" : (string) "",
+	  url->password ? url->password : (string) "",
+	  url->user ? (string) "@" : (string) "",
+	  url->machine && strchr(url->machine, ':') ? "[" : "",
+	  url->machine ? url->machine : (string) "",
+	  url->machine && strchr(url->machine, ':') ? "]" : "",
+	  url->port ? (string) ":" : (string) "",
+	  url->port ? url->port : (string) "",
+	  url->path ? url->path : (string) "",
+	  url->query ? "?" : (string) "",
+	  url->query ? url->query : (string) "",
+	  url->fragment ? (string) "#" : (string) "",
+	  url->fragment ? url->fragment : (string) "");
+  /* Instead of strchr() above, we could have an IPv6 flag. Necessary? */
+
+  return url;
+}
+
+/* URL_s_to_ascii -- use punycode and %-escaping to turn an IRI into a URI */
+EXPORT string URL_s_to_ascii(const conststring iri)
+{
+  URL x = URL_new(iri);
+  URL y = URL_to_ascii(x);
+  string result = newstring(y->full);
+  URL_dispose(y);
+  URL_dispose(x);
+  return result;
+}
